module mem(
        input           clk,            //reference clock
        input           rstn ,          //resetn, low effective

        input           en ,            //start to generating waves
        input [1:0]     sel ,           //waves selection

        input [7:0]     addr ,
        output          dout_en ,
        output [9:0]    dout);          //data out, 10bit width

   //data out from ROMs
   wire [9:0]           q_tri ;
   wire [9:0]           q_square ;
   wire [9:0]           q_cos ;

   //rom addr
   reg [1:0]            en_r ;
   always @(posedge clk or negedge rstn) begin
      if (!rstn) begin
         en_r   <= 2'b0 ;
      end
      else begin
         en_r   <= {en_r[0], en} ;         //delay one cycle for en
      end
   end
   assign dout      = en_r[1] ? (q_tri | q_square | q_cos) : 10'b0 ;
   assign dout_en   = en_r[1] ;

   //rom instiation
   cos_rom      u_cos_rom (
       .clk     (clk),
       .en      (en_r[0] & (sel == 2'b0)),    //sel = 0, sel cos wave
       .addr    (addr[7:0]),
       .q       (q_cos[9:0]));

   square_rom   u_square_rom (
       .clk     (clk),
       .en      (en_r[0] & sel == 2'b01),    //sel = 1, sel square wave
       .addr    (addr[7:0]),
       .q       (q_square[9:0]));

   tri_rom      u_tri_rom (
       .clk     (clk),
       .en      (en_r[0] & sel == 2'b10),    //sel = 2, sel triangle wave
       .addr    (addr[7:0]),
       .q       (q_tri[9:0]));

endmodule


 //square waves rom
module square_rom (
   input               clk,
   input               en,
   input [7:0]         addr,
   output reg [9:0]     q);

   always @(posedge clk) begin
      if (en) begin
         q <= { 10{(addr < 128)} };     //1 in first half cycle, and 0 in second half cycle
      end
      else begin
         q <= 'b0 ;
      end
   end

endmodule


 //triangle waves rom
module tri_rom (
   input               clk,
   input               en,
   input [7:0]         addr,
   output reg [9:0]     q);

   always @(posedge clk) begin
      if (en) begin
         if (addr < 128) begin
            q <= {addr[6:0], 3'b0};     //rising edge, addr -> 0x0, 0x3f
         end
         else begin
            q <= 10'h3ff - {addr[6:0], 3'b0} ;//- {1'b0, addr[6:0], 2'b00}; //falling edge
         end
      end
      else begin
         q <= 'b0 ;
      end
   end

endmodule



//Better use mem ip.
//This format is easy for simulation
module cos_rom (
   input               clk,
   input               en,
   input [7:0]         addr,
   output reg [9:0]     q);

   wire [8:0]           rom_t [0 : 64] ;
   //as the symmetry of cos function, just store 1/4 data of one cycle
   assign rom_t[0:64] = {
               511, 510, 510, 509, 508, 507, 505, 503,
               501, 498, 495, 492, 488, 485, 481, 476,
               472, 467, 461, 456, 450, 444, 438, 431,
               424, 417, 410, 402, 395, 386, 378, 370,
               361, 352, 343, 333, 324, 314, 304, 294,
               283, 273, 262, 251, 240, 229, 218, 207,
               195, 183, 172, 160, 148, 136, 124, 111,
               99 , 87 , 74 , 62 , 50 , 37 , 25 , 12 ,
               0 } ;

   always @(posedge clk) begin
      if (en) begin
         if (addr[7:6] == 2'b00 ) begin         //first quadrant, addr -> [0, 63]
            q <= rom_t[addr[5:0]] + 10'd512 ;
         end
         else if (addr[7:6] == 2'b01 ) begin    //second quadrant, addr -> [64, 127]
            q <= 10'd512 - rom_t[64-addr[5:0]] ;
         end
         else if (addr[7:6] == 2'b10 ) begin    //third quadrant, addr -> [128, 192]
            q <= 10'd512 - rom_t[addr[5:0]];
         end
         else begin     //4th quadrant, addr -> [193, 256]
            q <= 10'd512 + rom_t[64-addr[5:0]];
         end
      end
      else begin
         q <= 'b0 ;
      end
   end

endmodule